class Unit(object):
    '''
    Read / Write attributes
        job_id
        missed_count
        difficulty
        state
        data
        agreement

    Read-only attributes
        updated_at
        created_at
        judgments_count
        id
    '''
